use serde::Serialize;
use super::value::writer::Writer;

pub use super::error::Error;
type Result<T> = std::result::Result<T, Error>;

#[derive(Debug)]
pub struct Serializer<'a, W: Writer> {
    pub write: &'a mut W,
}

impl<'a, W: Writer> Serializer<'a, W> {
    pub fn new(write: &'a mut W) -> Self {
        Serializer { write }
    }
}

#[derive(Debug)]
pub struct SerializeCompound<'a, 'b, W: Writer> {
    ser: &'a mut Serializer<'b, W>,
    c: W::Compound,
}

#[derive(Debug)]
pub struct SerializeDictionary<'a, 'b, W: Writer> {
    ser: &'a mut Serializer<'b, W>,
    d: W::Dictionary,
    key_p: Option<W::KeyPointer>,
}

impl<'a, 'b, W: Writer> serde::Serializer for &'a mut Serializer<'b, W> {
    type Ok = W::Pointer;
    type Error = Error;
    type SerializeSeq = SerializeCompound<'a, 'b, W>;
    type SerializeTuple = SerializeCompound<'a, 'b, W>;
    type SerializeTupleStruct = SerializeCompound<'a, 'b, W>;
    type SerializeTupleVariant = SerializeCompound<'a, 'b, W>;
    type SerializeMap = SerializeDictionary<'a, 'b, W>;
    type SerializeStruct = SerializeCompound<'a, 'b, W>;
    type SerializeStructVariant = SerializeCompound<'a, 'b, W>;

    fn serialize_bool(self, v: bool) -> Result<Self::Ok> {
        Ok(self.write.write_bool(v)?)
    }

    fn serialize_i8(self, v: i8) -> Result<Self::Ok> {
        Ok(self.write.write_i8(v)?)
    }

    fn serialize_i16(self, v: i16) -> Result<Self::Ok> {
        Ok(self.write.write_i16(v)?)
    }

    fn serialize_i32(self, v: i32) -> Result<Self::Ok> {
        Ok(self.write.write_i32(v)?)
    }

    fn serialize_i64(self, v: i64) -> Result<Self::Ok> {
        Ok(self.write.write_i64(v)?)
    }

    fn serialize_u8(self, v: u8) -> Result<Self::Ok> {
        Ok(self.write.write_u8(v)?)
    }

    fn serialize_u16(self, v: u16) -> Result<Self::Ok> {
        Ok(self.write.write_u16(v)?)
    }

    fn serialize_u32(self, v: u32) -> Result<Self::Ok> {
        Ok(self.write.write_u32(v)?)
    }

    fn serialize_u64(self, v: u64) -> Result<Self::Ok> {
        Ok(self.write.write_u64(v)?)
    }

    fn serialize_f32(self, v: f32) -> Result<Self::Ok> {
        Ok(self.write.write_f32(v)?)
    }

    fn serialize_f64(self, v: f64) -> Result<Self::Ok> {
        Ok(self.write.write_f64(v)?)
    }

    fn serialize_char(self, v: char) -> Result<Self::Ok> {
        let mut c = self.write.start_record(1)?;
        let p = self.write.write_symbol("UnicodeScalar")?;
        self.write.extend_compound(&mut c, p)?;
        let p = self.write.write_u32(v as u32)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(self.write.end_compound(c)?)
    }

    fn serialize_str(self, v: &str) -> Result<Self::Ok> {
        Ok(self.write.write_string(v)?)
    }

    fn serialize_bytes(self, v: &[u8]) -> Result<Self::Ok> {
        Ok(self.write.write_bytes(v)?)
    }

    fn serialize_none(self) -> Result<Self::Ok> {
        let mut c = self.write.start_record(0)?;
        let p = self.write.write_symbol("None")?;
        self.write.extend_compound(&mut c, p)?;
        Ok(self.write.end_compound(c)?)
    }

    fn serialize_some<T: ?Sized>(self, v: &T) -> Result<Self::Ok> where T: Serialize {
        let mut c = self.write.start_record(1)?;
        let p = self.write.write_symbol("Some")?;
        self.write.extend_compound(&mut c, p)?;
        let p = v.serialize(&mut *self)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(self.write.end_compound(c)?)
    }

    fn serialize_unit(self) -> Result<Self::Ok> {
        let mut c = self.write.start_record(0)?;
        let p = self.write.write_symbol("tuple")?;
        self.write.extend_compound(&mut c, p)?;
        Ok(self.write.end_compound(c)?)
    }

    fn serialize_unit_struct(self, name: &'static str) -> Result<Self::Ok> {
        let mut c = self.write.start_record(0)?;
        let p = self.write.write_symbol(name)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(self.write.end_compound(c)?)
    }

    fn serialize_unit_variant(self,
                              _name: &'static str,
                              _variant: u32,
                              variant_name: &'static str) ->
        Result<Self::Ok>
    {
        let mut c = self.write.start_record(0)?;
        let p = self.write.write_symbol(variant_name)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(self.write.end_compound(c)?)
    }

    fn serialize_newtype_struct<T: ?Sized>(self, name: &'static str, value: &T) ->
        Result<Self::Ok> where T: Serialize
    {
        match super::value::magic::receive_output_value(name, value) {
            Some(v) => {
                Ok(self.write.write(&v)?)
            }
            None => {
                // TODO: This is apparently discouraged, and we should apparently just serialize `value`?
                let mut c = self.write.start_record(1)?;
                let p = self.write.write_symbol(name)?;
                self.write.extend_compound(&mut c, p)?;
                let p = value.serialize(&mut *self)?;
                self.write.extend_compound(&mut c, p)?;
                Ok(self.write.end_compound(c)?)
            }
        }
    }

    fn serialize_newtype_variant<T: ?Sized>(self,
                                            _name: &'static str,
                                            _variant: u32,
                                            variant_name: &'static str,
                                            value: &T) ->
        Result<Self::Ok> where T: Serialize
    {
        let mut c = self.write.start_record(1)?;
        let p = self.write.write_symbol(variant_name)?;
        self.write.extend_compound(&mut c, p)?;
        let p = value.serialize(&mut *self)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(self.write.end_compound(c)?)
    }

    fn serialize_seq(self, count: Option<usize>) -> Result<Self::SerializeSeq> {
        let c = match count {
            Some(n) => self.write.start_sequence(n)?,
            None => match self.write.stream_sequence()? {
                Some(c) => c,
                None => return Err(Error::StreamingSerializationUnsupported),
            }
        };
        Ok(SerializeCompound { ser: self, c })
    }

    fn serialize_tuple(self, count: usize) -> Result<Self::SerializeTuple> {
        let mut c = self.write.start_record(count)?;
        let p = self.write.write_symbol("tuple")?;
        self.write.extend_compound(&mut c, p)?;
        Ok(SerializeCompound { ser: self, c })
    }

    fn serialize_tuple_struct(self, name: &'static str, count: usize) ->
        Result<Self::SerializeTupleStruct>
    {
        let mut c = self.write.start_record(count)?;
        let p = self.write.write_symbol(name)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(SerializeCompound { ser: self, c })
    }

    fn serialize_tuple_variant(self,
                               _name: &'static str,
                               _variant: u32,
                               variant_name: &'static str,
                               count: usize) ->
        Result<Self::SerializeTupleVariant>
    {
        let mut c = self.write.start_record(count)?;
        let p = self.write.write_symbol(variant_name)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(SerializeCompound { ser: self, c })
    }

    fn serialize_map(self, count: Option<usize>) -> Result<Self::SerializeMap> {
        let d = match count {
            Some(n) => self.write.start_dictionary(n)?,
            None => match self.write.stream_dictionary()? {
                Some(d) => d,
                None => return Err(Error::StreamingSerializationUnsupported),
            }
        };
        Ok(SerializeDictionary { ser: self, d, key_p: None })
    }

    fn serialize_struct(self, name: &'static str, count: usize) -> Result<Self::SerializeStruct> {
        let mut c = self.write.start_record(count)?;
        let p = self.write.write_symbol(name)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(SerializeCompound { ser: self, c })
    }

    fn serialize_struct_variant(self,
                                _name: &'static str,
                                _variant: u32,
                                variant_name: &'static str,
                                count: usize) ->
        Result<Self::SerializeStructVariant>
    {
        let mut c = self.write.start_record(count)?;
        let p = self.write.write_symbol(variant_name)?;
        self.write.extend_compound(&mut c, p)?;
        Ok(SerializeCompound { ser: self, c })
    }
}

impl<'a, 'b, W: Writer> serde::ser::SerializeMap for SerializeDictionary<'a, 'b, W> {
    type Ok = W::Pointer;
    type Error = Error;

    fn serialize_key<T: ?Sized>(&mut self, key: &T) -> Result<()> where T: Serialize {
        let kp1 = key.serialize(&mut *self.ser)?;
        self.key_p = Some(self.ser.write.extend_dictionary_key(&mut self.d, kp1)?);
        Ok(())
    }

    fn serialize_value<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        let vp1 = value.serialize(&mut *self.ser)?;
        Ok(self.ser.write.extend_dictionary_value(&mut self.d, self.key_p.take().unwrap(), vp1)?)
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(self.ser.write.end_dictionary(self.d)?)
    }
}

impl<'a, 'b, W: Writer> serde::ser::SerializeStruct for SerializeCompound<'a, 'b, W> {
    type Ok = W::Pointer;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        let p = value.serialize(&mut *self.ser)?;
        Ok(self.ser.write.extend_compound(&mut self.c, p)?)
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(self.ser.write.end_compound(self.c)?)
    }
}

impl<'a, 'b, W: Writer> serde::ser::SerializeStructVariant for SerializeCompound<'a, 'b, W> {
    type Ok = W::Pointer;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        let p = value.serialize(&mut *self.ser)?;
        Ok(self.ser.write.extend_compound(&mut self.c, p)?)
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(self.ser.write.end_compound(self.c)?)
    }
}

impl<'a, 'b, W: Writer> serde::ser::SerializeTuple for SerializeCompound<'a, 'b, W> {
    type Ok = W::Pointer;
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        let p = value.serialize(&mut *self.ser)?;
        Ok(self.ser.write.extend_compound(&mut self.c, p)?)
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(self.ser.write.end_compound(self.c)?)
    }
}

impl<'a, 'b, W: Writer> serde::ser::SerializeTupleStruct for SerializeCompound<'a, 'b, W> {
    type Ok = W::Pointer;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        let p = value.serialize(&mut *self.ser)?;
        Ok(self.ser.write.extend_compound(&mut self.c, p)?)
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(self.ser.write.end_compound(self.c)?)
    }
}

impl<'a, 'b, W: Writer> serde::ser::SerializeTupleVariant for SerializeCompound<'a, 'b, W> {
    type Ok = W::Pointer;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        let p = value.serialize(&mut *self.ser)?;
        Ok(self.ser.write.extend_compound(&mut self.c, p)?)
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(self.ser.write.end_compound(self.c)?)
    }
}

impl<'a, 'b, W: Writer> serde::ser::SerializeSeq for SerializeCompound<'a, 'b, W> {
    type Ok = W::Pointer;
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        let p = value.serialize(&mut *self.ser)?;
        Ok(self.ser.write.extend_compound(&mut self.c, p)?)
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(self.ser.write.end_compound(self.c)?)
    }
}

pub fn to_writer<W: Writer, T: Serialize>(write: &mut W, value: &T) -> Result<W::Pointer> {
    let mut ser: Serializer<'_, W> = Serializer::new(write);
    value.serialize(&mut ser)
}
